module WeatherModel
  module Storer
    require 'active_record'
    require 'ohm'
    require "ohm/expire"

    class Mysql < ::ActiveRecord::Base
      after_save :update_ohm
      def self.use_settings settings
        self.table_name = settings[:table_name]
        set_options settings
        define_ohm settings
        define_factory settings
      end

      def self.set_options settings
        [:has_many, :belongs_to, :json_column].each do |option|
          attrs = settings[:options][option]
          attrs.is_a?(Array) ?
            attrs.each{ |attr| method(option).call(attr) if attr } :
            method(option).call(attrs) if attrs
        end
      end


      def self.define_ohm settings
        const_set('Ohm', Class.new(SourceOhm)).use_settings(settings)
      end

      def self.define_factory settings
        model_name = self
        ::FactoryGirl.define do
          factory settings[:factory_name], class: model_name do
            settings[:factory].each_pair{ |key, val| send(key, val) }
          end
        end
      end

      def get_ohm
        self.class.const_get('Ohm')
      end

      # Forecast -> :forecasts , haven't tested
      def self.has_many name
        super name.to_s.underscore.pluralize.to_sym
      end
      # Forecast -> :forecast
      def self.belongs_to name
        super name.to_s.underscore.to_sym
      end

      def self.json_column column
        serialize column 
      end


      class SourceOhm < ::Ohm::Model
        include ::Ohm::Expire
        
        def self.use_settings settings

          attribute :mysql_id
          attribute :updated_at
          index :mysql_id
          unique :mysql_id

          settings[:schema].each_pair { |attr, type|
            attribute attr, type_proc(type)
          }
          set_options settings
        end

        def self.set_options settings
          [:index, :unique, :expire, :has_many, :belongs_to, :json_column].each do |option|
            attrs = settings[:options][option]
            attrs.is_a?(Array) ?
              attrs.each{ |attr| method(option).call(attr) if attr } :
              method(option).call(attrs) if attrs
          end
        end

        def self.all
          super.select{ |x| x.attributes.empty?}.map(&:delete)
          super
        end
        
        def self.type_proc type
          case type
          when :integer
            proc { |val| val && val.to_i }
          when :float
            proc { |val| val && val.to_f }
          when :datetime
            proc { |val| val.is_a?(String) ? Time.parse(val) : val}
          when :json
            proc { |value|
              value.is_a?(String) ?
                JSON.parse(value.gsub('=>', ':')) :
                value
            }
          end
        end

        # ohm-expire 只处理了create
        # 而且 只把 attributes 的值给 TTL了。。。
        def update attributes
          obj = super(attributes)
          obj.update_ttl self.class.instance_variable_get(:@expire)
        end

        def to_h
          keys = attributes.keys - [:mysql_id, :updated_at]
          Hash[keys.zip keys.map { |column_name| send(column_name) }]
        end


        private

          def self.has_many his_son # input like :Post
            # collection :posts, :Post
            collection(his_son.to_s.underscore.pluralize.to_sym, his_son) if his_son && SourceSchema.const_defined?(his_son)
          end

          def self.belongs_to his_dad # input like :Post
            # reference :post, :Post
            reference(his_dad.to_s.underscore.to_sym, his_dad ) if his_dad && SourceSchema.const_defined?(his_dad)
          end

          def self.json_column column
            attribute column, proc { |value|
              value.is_a?(String) ?
                JSON.parse(value.gsub('=>', ':')) :
                value
            }
          end
        end


    private
      def update_ohm
        ohm_attrs = self.attributes.except('created_at', 'id').merge(mysql_id: id)
        # find 找不到 会自动生成
        ohm_obj = get_ohm.find(mysql_id: id).first.try(:update, ohm_attrs) || get_ohm.create(ohm_attrs)
        # rescue  Ohm::UniqueIndexViolation mysql_id should be unique
      end
    end   
  end
end
